"""Run checks."""

import asyncio
import datetime as dt
import logging
import signal
import types
from functools import partial
from threading import Event

import click
from regal.exceptions import NssetDoesNotExist
from setapp import Verbosity
from sqlalchemy import delete, null, update

import dnscheck
from dnscheck.db import DomainCheck, NssetCheck, atomic
from dnscheck.process import domain_checks, nsset_checks, nssets
from dnscheck.settings import APP
from dnscheck.utils import StopProcess, async_to_sync, defer_errors, print, run_unless_stopped, verbosity_options

_LOGGER = logging.getLogger(__name__)

_DATETIME_FORMATS = ("%Y-%m-%dT%H:%M:%S%z", "%Y-%m-%d %H:%M:%S%z")


def _signal_handler(signum: int, frame: types.FrameType | None, *, stop_event: Event) -> None:  # pragma: no cover
    """Capture signal and set stop event."""
    signal_name = signal.Signals(signum).name
    print(f"Captured signal {signal_name}, stopping dnscheck daemon", level=Verbosity.DETAIL)
    _LOGGER.info("Captured signal %s, stopping dnscheck daemon", signal_name)
    stop_event.set()


@click.group()
@click.help_option("-h", "--help")
@click.version_option(dnscheck.__version__, "-V", "--version", message="fred-dnscheck %(version)s")
@click.option("-c", "--config", type=click.Path(exists=True, dir_okay=False), help="Config file.")
@verbosity_options
def cli(config: str) -> None:
    """DNScheck CLI."""
    APP.load(config)


@cli.command()
@verbosity_options
@click.option(
    "--batch-size", default=1000, show_default=True, help="How many nssets should be processed simultaneously."
)
@async_to_sync
async def sync(batch_size: int) -> None:
    """Sync nssets with domain registry."""
    await nssets.sync(batch_size=batch_size)


@cli.command()
@verbosity_options
@click.option(
    "--loop-delay",
    envvar="DNSCHECK_LOOP_DELAY",
    default=5,
    show_default=True,
    help="Delay in seconds between processing loop runs.",
)
@async_to_sync
async def run(loop_delay: int) -> None:
    """Run main process loop."""
    print("Starting dnscheck daemon")
    _LOGGER.info("Starting dnscheck daemon")
    stop_event = Event()
    signal.signal(signal.SIGTERM, partial(_signal_handler, stop_event=stop_event))

    _run = run_unless_stopped(defer_errors, stop_event=stop_event)
    try:
        while True:
            await _run(domain_checks.finish)()
            await _run(nsset_checks.start)()
            await _run(domain_checks.start)()
            await _run(nsset_checks.finish)()
            await _run(asyncio.sleep)(loop_delay)
    except KeyboardInterrupt:
        print("Stopping dnscheck daemon")
        _LOGGER.info("Stopping dnscheck daemon")
    except StopProcess:
        print("Dnscheck daemon has been gracefully stopped")
        _LOGGER.info("Dnscheck daemon has been gracefully stopped")


@cli.command()
@verbosity_options
@click.option("-n", "--dry-run", is_flag=True)
@click.option(
    "--before",
    type=click.DateTime(formats=_DATETIME_FORMATS),
    required=True,
    help="Domain check results older than this timezone-aware datetime will be cleaned.",
)
def clean_results(before: dt.datetime, dry_run: bool) -> None:
    """Clean old domain check results.

    The results are replaced with NULL.
    The domain checks themselves are preserved.
    """
    if dry_run:
        print("[bold yellow]Dry run mode is active, nothing will be changed[/bold yellow]")

    with atomic() as db:
        result = db.execute(
            update(DomainCheck)
            .where(DomainCheck.finished_at < before, DomainCheck.results.is_not(null()))
            .values(results=null())
        )
        print(f"Cleaned domain check results: {result.rowcount}")
        if dry_run:
            db.rollback()


@cli.command()
@verbosity_options
@click.option("-n", "--dry-run", is_flag=True)
@click.option(
    "--before",
    type=click.DateTime(formats=_DATETIME_FORMATS),
    required=True,
    help="Checks older than this timezone-aware datetime will be cleaned.",
)
def clean_checks(before: dt.datetime, dry_run: bool) -> None:
    """Clean old nsset checks and domain checks."""
    if dry_run:
        print("[bold yellow]Dry run mode is active, nothing will be changed[/bold yellow]")

    with atomic() as db:
        result = db.execute(delete(NssetCheck).where(NssetCheck.finished_at < before))
        print(f"Deleted nsset checks: {result.rowcount}")
        if dry_run:
            db.rollback()


@cli.command()
@verbosity_options
@click.argument("handle", nargs=-1)
@async_to_sync
async def start_check(handle: tuple[str]) -> None:
    """Start NSSET checks."""
    for nsset_handle in handle:
        try:
            nsset_id = await APP.nsset_client.get_nsset_id(nsset_handle)
        except NssetDoesNotExist:
            print(f"[bold red]Nsset {nsset_handle} does not exist[/bold red]")
            continue
        await nsset_checks.start_one(nsset_id)
        print(f"Started nsset check of {nsset_handle}", level=Verbosity.DETAIL)
