"""Process nsset functions."""

import datetime as dt
import logging

from asyncstdlib.itertools import batched
from rich.progress import Progress, SpinnerColumn, TextColumn
from setapp import Verbosity
from sqlalchemy import delete, insert, text, update

from dnscheck.db import Nsset, atomic
from dnscheck.settings import APP
from dnscheck.utils import print

_LOGGER = logging.getLogger(__name__)

_LOCK_KEY = hash("dnscheck_sync_lock")


async def sync(batch_size: int = 1000) -> None:
    """Sync nssets with registry.

    Args:
        batch_size: Size of nssets chunk that are processed together.
    """
    _LOGGER.info("Running nsset sync")
    sync_started_at = dt.datetime.now(tz=dt.UTC)
    summary = {"insert": 0, "update": 0, "delete": 0}

    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        transient=True,
        console=APP.console,
        disable=(APP.verbosity == Verbosity.SILENT),
    ) as progress:
        progress.add_task("Syncing...", total=None)
        with atomic() as sess:
            if sess.connection().dialect.name == "postgresql":
                # xact locks are transaction scoped, there's no explicit release
                lock_stmt = text("SELECT pg_try_advisory_xact_lock(:key)").params(key=_LOCK_KEY)
                if not sess.execute(lock_stmt).scalar_one():
                    print("[red bold]Another nsset sync is already running. Aborting.[/red bold]")
                    _LOGGER.warning("Another nsset sync is already running")
                    return

            async for nssets in batched(APP.nsset_client.list(exclude_not_linked=True), batch_size):
                # First update all the existing nssets
                updated_ids = set(
                    sess.execute(
                        update(Nsset)
                        .values(last_synced=dt.datetime.now(tz=dt.UTC))
                        .where(Nsset.nsset_id.in_(n.id for n in nssets))
                        .returning(Nsset.nsset_id)
                    ).scalars()
                )
                summary["update"] += len(updated_ids)
                print(f"Updated {len(updated_ids)} existing nssets", level=Verbosity.DEBUG)
                _LOGGER.info("Updated %s existing nssets", len(updated_ids))

                # Then insert all the new nssets
                new_nssets = [{"nsset_id": n.id, "handle": n.handle} for n in nssets if n.id not in updated_ids]
                if new_nssets:
                    sess.execute(insert(Nsset).values(new_nssets))
                    summary["insert"] += len(new_nssets)
                    print(f"Inserted {len(new_nssets)} existing nssets", level=Verbosity.DEBUG)
                    _LOGGER.info("Inserted %s new nssets", len(new_nssets))

            # Last, delete all nssets that don't exist anymore
            result = sess.execute(delete(Nsset).where(Nsset.last_synced < sync_started_at))
            summary["delete"] += result.rowcount
            print(f"Deleted {result.rowcount} old nssets", level=Verbosity.DEBUG)
            _LOGGER.info("Deleted %s old nssets", result.rowcount)

    print("[bold]Summary:[/bold] {insert} inserted, {update} updated, {delete} deleted".format(**summary))
