import datetime as dt
from ipaddress import IPv4Address, IPv6Address
from unittest.mock import call

import pytest
from asyncstdlib import iter as aiter
from fred_types import ContactRef, DomainId, DomainRef, NssetHistoryId, NssetId
from hermes import Email, Reference
from hermes.client import SendFailed
from pytest_mock import MockFixture
from regal import DnsHost, Nsset as NssetSchema
from regal.exceptions import ContactDoesNotExist, NssetDoesNotExist
from sqlalchemy import func, select
from testfixtures import LogCapture, SequenceComparison as S
from time_machine import TimeMachineFixture

from dnscheck.constants import (
    DomainCheckStatus,
    EmailType,
    MessageTemplate,
    NssetCheckStatus,
    ReferenceType,
    TestResultLevel,
)
from dnscheck.db import DomainCheck, NssetCheck, atomic
from dnscheck.process import nsset_checks
from dnscheck.settings import APP
from dnscheck.tests.utils import ModelFactory, RegistryFactory, override_settings


@pytest.fixture(autouse=True)
def clear_cache():
    nsset_checks._CACHE.clear()


class ListNssetCheckCandidatesTest:
    @pytest.fixture(autouse=True)
    def setup(self, db, time_machine: TimeMachineFixture):
        now = dt.datetime(2000, 1, 24, tzinfo=dt.UTC)
        time_machine.move_to(now)
        dt.datetime.now()  # start the time machine
        self.mf = mf = ModelFactory()

        with atomic() as sess:
            sess.add_all(
                [
                    mf.nsset(1),
                    mf.nsset(2),
                    mf.nsset(3),
                    mf.nsset(4),
                    mf.nsset(5),
                    mf.nsset(6),
                    mf.nsset_check(
                        1,
                        started_at=now - dt.timedelta(days=30),
                        finished_at=now - dt.timedelta(days=29),
                        status=NssetCheckStatus.FINISHED,
                    ),
                    mf.nsset_check(1, started_at=now - dt.timedelta(days=1)),
                    mf.nsset_check(2, started_at=now - dt.timedelta(days=30)),
                    mf.nsset_check(3, started_at=now - dt.timedelta(days=2), finished_at=now - dt.timedelta(days=1)),
                    mf.nsset_check(4, started_at=now - dt.timedelta(days=30), finished_at=now - dt.timedelta(days=29)),
                    mf.nsset_check(5, started_at=now - dt.timedelta(days=20), finished_at=now - dt.timedelta(days=19)),
                ]
            )

    @override_settings(min_check_interval=dt.timedelta(days=10), check_once_per_month=True)
    def test_once_per_month(self, time_machine: TimeMachineFixture):
        nsset_ids = nsset_checks.list_candidates()

        assert list(nsset_ids) == [
            self.mf.nsset(6).nsset_id,
            self.mf.nsset(4).nsset_id,
        ]
        # 6 has never been checked
        # 4 has been checked 29 days ago, in a previous month
        #
        # 1 and 2 have checks that are still running
        # 3 has been checked 1 day ago, which is too recent
        # 5 has been checked 19 days ago, but in the current month

    @override_settings(min_check_interval=dt.timedelta(days=10), check_once_per_month=False)
    def test_more_per_month(self, time_machine: TimeMachineFixture):
        nsset_ids = nsset_checks.list_candidates()

        assert list(nsset_ids) == [
            self.mf.nsset(6).nsset_id,
            self.mf.nsset(4).nsset_id,
            self.mf.nsset(5).nsset_id,
        ]
        # 6 has never been checked
        # 4 has been checked 29 days ago
        # 5 has been checked 19 days ago
        #
        # 1 and 2 have checks that are still running
        # 3 has been checked 1 day ago, which is too recent


class StartNssetChecksTest:
    @pytest.fixture(autouse=True)
    def setup(self, mocker: MockFixture):
        mf = self.mf = ModelFactory()

        self.mock_list = mocker.patch(
            "dnscheck.process.nsset_checks.list_candidates",
            return_value=iter([NssetId(mf.nsset(id).nsset_id) for id in (1, 2, 3)]),
        )
        self.mock_start = mocker.patch("dnscheck.process.nsset_checks.start_one", side_effect=[8, 5, 1])

    @pytest.fixture
    def create_domain_checks(self, db):
        with atomic() as sess:
            sess.add_all(
                [
                    self.mf.nsset(42),
                    self.mf.nsset_check(42),
                    self.mf.domain_check("1.example.org"),
                    self.mf.domain_check("2.example.org"),
                ]
            )

    async def test_start(self, db, mocker: MockFixture):
        # Start adds enough nsset checks to fill domain check quota ten times
        self.mock_start = mocker.patch("dnscheck.process.nsset_checks.start_one", side_effect=[50, 50, 1])
        with override_settings(max_domain_checks=10):
            await nsset_checks.start()

        assert self.mock_list.mock_calls == [call()]
        assert self.mock_start.mock_calls == [call(self.mf.nsset(1).nsset_id), call(self.mf.nsset(2).nsset_id)]

    async def test_start_no_candidates(self, db, create_domain_checks, mocker: MockFixture):
        self.mock_list = mocker.patch("dnscheck.process.nsset_checks.list_candidates", return_value=iter([]))

        with override_settings(max_domain_checks=10):
            await nsset_checks.start()

        assert self.mock_list.mock_calls == [call()]
        assert self.mock_start.mock_calls == []

    async def test_start_no_candidates_cache(self, db, create_domain_checks, mocker: MockFixture):
        self.mock_list = mocker.patch("dnscheck.process.nsset_checks.list_candidates", return_value=iter([]))

        with override_settings(max_domain_checks=10):
            await nsset_checks.start()
            await nsset_checks.start()

        # list_candidates is called only once, empty result is cached
        assert self.mock_list.mock_calls == [call()]
        assert self.mock_start.mock_calls == []

    async def test_running_checks(self, db, create_domain_checks, mocker: MockFixture):
        # Start adds enough nsset checks to fill domain check quota ten times
        self.mock_start = mocker.patch("dnscheck.process.nsset_checks.start_one", side_effect=[98, 1, 1])
        with override_settings(max_domain_checks=10):
            await nsset_checks.start()

        assert self.mock_list.mock_calls == [call()]
        assert self.mock_start.mock_calls == [call(self.mf.nsset(1).nsset_id)]

    async def test_too_many_running_checks(self, db, create_domain_checks):
        with override_settings(max_domain_checks=2):
            await nsset_checks.start()

        assert self.mock_list.mock_calls == []  # expensive list operation is not called
        assert self.mock_start.mock_calls == []


@override_settings(domain_zones=[])
async def test_start_nsset_check(db, mocker: MockFixture):
    mocker.patch.object(
        APP.nsset_client,
        "get_nsset_info",
        return_value=NssetSchema(
            nsset_id=NssetId("NSS-1"),
            nsset_handle="NSSET",
            nsset_history_id=NssetHistoryId("NSS-2"),
            dns_hosts=[
                DnsHost(fqdn="ns1.example.org", ip_addresses=[]),
                DnsHost(fqdn="ns2.example.org", ip_addresses=[IPv4Address("127.0.0.1"), IPv6Address("::1")]),
            ],
        ),
    )
    mocker.patch.object(
        APP.domain_client,
        "list_by_nsset",
        return_value=aiter(
            [
                DomainRef(id=DomainId("DOMAIN-1"), fqdn="domain1.example.org"),
                DomainRef(id=DomainId("DOMAIN-2"), fqdn="domain2.example.org"),
                DomainRef(id=DomainId("DOMAIN-3"), fqdn="domain3.example.org"),
            ]
        ),
    )

    await nsset_checks.start_one(NssetId("NSS-1"))

    with atomic() as sess:
        check = sess.execute(select(NssetCheck)).scalar_one()
        assert check.nsset_id == "NSS-1"
        assert check.nsset_history_id == "NSS-2"
        assert check.handle == "NSSET"
        assert check.nameservers == [
            {"fqdn": "ns1.example.org", "ip_addresses": []},
            {"fqdn": "ns2.example.org", "ip_addresses": ["127.0.0.1", "::1"]},
        ]
        domain_checks = sess.execute(select(DomainCheck)).scalars().all()
        assert [dc.domain for dc in domain_checks] == [
            "domain1.example.org",
            "domain2.example.org",
            "domain3.example.org",
        ]
        assert all(dc.nsset_check == check for dc in domain_checks)


@override_settings(domain_zones=[])
async def test_start_nsset_check_no_domains(db, mocker: MockFixture):
    mocker.patch.object(
        APP.nsset_client,
        "get_nsset_info",
        return_value=NssetSchema(
            nsset_id=NssetId("NSS-1"),
            nsset_handle="NSSET",
            nsset_history_id=NssetHistoryId("NSS-2"),
            dns_hosts=[
                DnsHost(fqdn="ns1.example.org", ip_addresses=[]),
                DnsHost(fqdn="ns2.example.org", ip_addresses=[IPv4Address("127.0.0.1"), IPv6Address("::1")]),
            ],
        ),
    )
    mocker.patch.object(APP.domain_client, "list_by_nsset", return_value=aiter([]))

    await nsset_checks.start_one(NssetId("NSS-1"))

    with atomic() as sess:
        nsset_check = sess.execute(select(NssetCheck)).scalar_one()
        assert nsset_check.status == NssetCheckStatus.FINISHED
        assert sess.execute(select(func.count(DomainCheck.id))).scalar_one() == 0


@override_settings(domain_zones=["com", "org"])
async def test_start_one_with_zones(db, mocker: MockFixture):
    mocker.patch.object(
        APP.nsset_client,
        "get_nsset_info",
        return_value=NssetSchema(
            nsset_id=NssetId("NSS-1"),
            nsset_handle="NSSET",
            nsset_history_id=NssetHistoryId("NSS-2"),
            dns_hosts=[
                DnsHost(fqdn="ns1.example.org", ip_addresses=[]),
                DnsHost(fqdn="ns2.example.org", ip_addresses=[IPv4Address("127.0.0.1"), IPv6Address("::1")]),
            ],
        ),
    )
    mocker.patch.object(
        APP.domain_client,
        "list_by_nsset",
        return_value=aiter(
            [
                DomainRef(id=DomainId("DOMAIN-1"), fqdn="domain1.example.com"),
                DomainRef(id=DomainId("DOMAIN-2"), fqdn="domain2.example.org"),
                DomainRef(id=DomainId("DOMAIN-3"), fqdn="domain3.example.net"),
            ]
        ),
    )

    await nsset_checks.start_one(NssetId("NSS-1"))

    with atomic() as sess:
        check = sess.execute(select(NssetCheck)).scalar_one()
        assert check.nsset_id == "NSS-1"
        assert check.nsset_history_id == "NSS-2"
        assert check.handle == "NSSET"
        assert check.nameservers == [
            {"fqdn": "ns1.example.org", "ip_addresses": []},
            {"fqdn": "ns2.example.org", "ip_addresses": ["127.0.0.1", "::1"]},
        ]
        domain_checks = sess.execute(select(DomainCheck)).scalars().all()
        assert [dc.domain for dc in domain_checks] == [
            "domain1.example.com",
            "domain2.example.org",
            # .net is missing, because it's not in allowed zones
        ]
        assert all(dc.nsset_check == check for dc in domain_checks)


@override_settings(domain_zones=["arpa"])
async def test_start_one_with_zones_no_domains(db, mocker: MockFixture):
    mocker.patch.object(
        APP.nsset_client,
        "get_nsset_info",
        return_value=NssetSchema(
            nsset_id=NssetId("NSS-1"),
            nsset_handle="NSSET",
            nsset_history_id=NssetHistoryId("NSS-2"),
            dns_hosts=[
                DnsHost(fqdn="ns1.example.org", ip_addresses=[]),
                DnsHost(fqdn="ns2.example.org", ip_addresses=[IPv4Address("127.0.0.1"), IPv6Address("::1")]),
            ],
        ),
    )
    mocker.patch.object(
        APP.domain_client,
        "list_by_nsset",
        return_value=aiter(
            [
                DomainRef(id=DomainId("DOMAIN-1"), fqdn="domain1.example.com"),
                DomainRef(id=DomainId("DOMAIN-2"), fqdn="domain2.example.org"),
                DomainRef(id=DomainId("DOMAIN-3"), fqdn="domain3.example.net"),
            ]
        ),
    )

    await nsset_checks.start_one(NssetId("NSS-1"))

    with atomic() as sess:
        nsset_check = sess.execute(select(NssetCheck)).scalar_one()
        assert nsset_check.status == NssetCheckStatus.FINISHED
        domain_checks = sess.execute(select(DomainCheck)).scalars().all()
        assert not domain_checks


async def test_start_one_does_not_exist(db, mocker: MockFixture):
    mocker.patch.object(APP.nsset_client, "get_nsset_info", side_effect=NssetDoesNotExist)

    await nsset_checks.start_one(NssetId("NSS-1"))

    with atomic() as sess:
        assert sess.execute(select(func.count(NssetCheck.id))).scalar_one() == 0
        assert sess.execute(select(func.count(DomainCheck.id))).scalar_one() == 0


async def test_start_one_index_violation(db, mocker: MockFixture):
    mocker.patch.object(
        APP.nsset_client,
        "get_nsset_info",
        return_value=NssetSchema(
            nsset_id=NssetId("NSS-1"),
            nsset_handle="NSSET",
            nsset_history_id=NssetHistoryId("NSS-2"),
            dns_hosts=[
                DnsHost(fqdn="ns1.example.org", ip_addresses=[]),
                DnsHost(fqdn="ns2.example.org", ip_addresses=[IPv4Address("127.0.0.1"), IPv6Address("::1")]),
            ],
        ),
    )
    domains = [
        DomainRef(id=DomainId("DOMAIN-1"), fqdn="domain1.example.org"),
        DomainRef(id=DomainId("DOMAIN-2"), fqdn="domain2.example.org"),
        DomainRef(id=DomainId("DOMAIN-3"), fqdn="domain3.example.org"),
    ]
    mocker.patch.object(
        APP.domain_client,
        "list_by_nsset",
        side_effect=[aiter(domains), aiter(domains)],
    )

    assert await nsset_checks.start_one(NssetId("NSS-1")) == 3
    assert await nsset_checks.start_one(NssetId("NSS-1")) == 0

    with atomic() as sess:
        assert sess.execute(select(func.count(NssetCheck.id))).scalar_one() == 1
        assert sess.execute(select(func.count(DomainCheck.id))).scalar_one() == 3


class FinishTest:
    @pytest.fixture(autouse=True)
    def setup(self, db):
        mf = ModelFactory()
        with atomic() as sess:
            sess.add_all(
                [
                    # all checks have finished
                    mf.nsset_check(1, status=NssetCheckStatus.STARTED),
                    mf.domain_check("1.example.org", status=DomainCheckStatus.FINISHED),
                    # one domain check is still running
                    mf.nsset_check(2, status=NssetCheckStatus.STARTED),
                    mf.domain_check("2.example.org", status=DomainCheckStatus.STARTED),
                    # there are no domain checks
                    mf.nsset_check(3, status=NssetCheckStatus.STARTED),
                ]
            )

    async def test_ok(db, mocker: MockFixture):
        mock_finish_one = mocker.patch("dnscheck.process.nsset_checks.finish_one")

        await nsset_checks.finish()

        assert mock_finish_one.mock_calls == S(call(1), call(3), ordered=False)

    async def test_unexpected_error(db, mocker: MockFixture):
        mock_finish_one = mocker.patch(
            "dnscheck.process.nsset_checks.finish_one", side_effect=[ZeroDivisionError, None]
        )

        await nsset_checks.finish()

        assert mock_finish_one.mock_calls == S(call(1), call(3), ordered=False)


class FinishOneTest:
    @pytest.fixture(autouse=True)
    def setup(self, db, mocker: MockFixture):
        mf = ModelFactory()

        with atomic() as sess:
            sess.add_all(
                [
                    nsset_check := mf.nsset_check(1, status=NssetCheckStatus.STARTED),
                    dc1 := mf.domain_check(
                        "1.example.org", status=DomainCheckStatus.FINISHED, summary={TestResultLevel.CRITICAL: 1}
                    ),
                    dc2 := mf.domain_check(
                        "2.example.org", status=DomainCheckStatus.FINISHED, summary={TestResultLevel.ERROR: 2}
                    ),
                    dc3 := mf.domain_check(
                        "3.example.org", status=DomainCheckStatus.FINISHED, summary={TestResultLevel.WARNING: 3}
                    ),
                ]
            )
            self.nsset_check = nsset_check
            self.domain_check1 = dc1
            self.domain_check2 = dc2
            self.domain_check3 = dc3

        rf = RegistryFactory()
        self.contact1 = rf.contact(1, emails=["rimmer@example.com"])
        self.contact2 = rf.contact(2, emails=["smeghead@example.com"])
        self.nsset = rf.nsset(
            1,
            technical_contacts=[
                ContactRef(id=self.contact1.contact_id, handle=self.contact1.contact_handle),
                ContactRef(id=self.contact2.contact_id, handle=self.contact2.contact_handle),
            ],
        )

    async def test_no_such_check(self, db, mocker: MockFixture):
        nsset_mock = mocker.patch.object(APP.nsset_client, "get_nsset_info", return_value=self.nsset)

        with atomic() as sess:
            sess.add(self.nsset_check)
            self.nsset_check.status = NssetCheckStatus.FINISHED

        with LogCapture() as logs:
            await nsset_checks.finish_one(1)

        assert nsset_mock.mock_calls == []
        logs.check()

    async def test_emails_disabled(self, db, mocker: MockFixture):
        self.nsset.technical_check_level = 2
        nsset_mock = mocker.patch.object(APP.nsset_client, "get_nsset_info", return_value=self.nsset)
        contact_mock = mocker.patch.object(
            APP.contact_client, "get_contact_info", side_effect=[self.contact1, self.contact2]
        )

        with LogCapture() as logs:
            await nsset_checks.finish_one(1)

        with atomic() as sess:
            sess.add(self.nsset_check)

            assert self.nsset_check.status == NssetCheckStatus.FINISHED
            assert nsset_mock.mock_calls == [call(self.nsset.nsset_id)]
            assert contact_mock.mock_calls == []

        logs.check(
            (
                "dnscheck.process.nsset_checks",
                "INFO",
                f"Would send email notifications about nsset {self.nsset.nsset_id}, but emails are disabled.",
            )
        )

    @override_settings(messenger_netloc="localhost:9999")
    async def test_check_ok(self, db, mocker: MockFixture):
        self.nsset.technical_check_level = 1
        nsset_mock = mocker.patch.object(APP.nsset_client, "get_nsset_info", return_value=self.nsset)
        contact_mock = mocker.patch.object(
            APP.contact_client, "get_contact_info", side_effect=[self.contact1, self.contact2]
        )
        with atomic() as sess:
            sess.delete(self.domain_check1)

        with LogCapture() as logs:
            await nsset_checks.finish_one(1)

        with atomic() as sess:
            sess.add(self.nsset_check)

            assert self.nsset_check.status == NssetCheckStatus.FINISHED
            assert nsset_mock.mock_calls == [call(self.nsset.nsset_id)]
            assert contact_mock.mock_calls == []

        logs.check()

    @override_settings(messenger_netloc="localhost:9999")
    async def test_level_disabled(self, db, mocker: MockFixture):
        self.nsset.technical_check_level = 0
        nsset_mock = mocker.patch.object(APP.nsset_client, "get_nsset_info", return_value=self.nsset)
        contact_mock = mocker.patch.object(
            APP.contact_client, "get_contact_info", side_effect=[self.contact1, self.contact2]
        )

        await nsset_checks.finish_one(1)

        with atomic() as sess:
            sess.add(self.nsset_check)

            assert self.nsset_check.status == NssetCheckStatus.FINISHED
            assert nsset_mock.mock_calls == [call(self.nsset.nsset_id)]
            assert contact_mock.mock_calls == []

    @override_settings(messenger_netloc="localhost:9999")
    async def test_level_critical_no_domains(self, db, mocker: MockFixture):
        # NOTE: This can happen when nsset has some domains, but all are eliminated
        # because of outzone condition or some other reason.
        self.nsset.technical_check_level = 1
        nsset_mock = mocker.patch.object(APP.nsset_client, "get_nsset_info", return_value=self.nsset)
        contact_mock = mocker.patch.object(
            APP.contact_client, "get_contact_info", side_effect=[self.contact1, self.contact2]
        )
        email_mock = mocker.patch.object(APP.email_client, "send")
        with atomic() as sess:
            sess.delete(self.domain_check1)
            sess.delete(self.domain_check2)
            sess.delete(self.domain_check3)

        await nsset_checks.finish_one(1)

        with atomic() as sess:
            sess.add(self.nsset_check)

            assert self.nsset_check.status == NssetCheckStatus.FINISHED
            assert nsset_mock.mock_calls == [call(self.nsset.nsset_id)]
            assert contact_mock.mock_calls == []
            email_mock.assert_not_called()

    @override_settings(messenger_netloc="localhost:9999")
    async def test_level_critical(self, db, mocker: MockFixture):
        self.nsset.technical_check_level = 1
        nsset_mock = mocker.patch.object(APP.nsset_client, "get_nsset_info", return_value=self.nsset)
        contact_mock = mocker.patch.object(
            APP.contact_client, "get_contact_info", side_effect=[self.contact1, self.contact2]
        )
        email_mock = mocker.patch.object(APP.email_client, "send", side_effect=["M1", "M2"])

        with LogCapture() as logs:
            await nsset_checks.finish_one(1)

        with atomic() as sess:
            sess.add_all([self.nsset_check, self.domain_check1])

            assert self.nsset_check.status == NssetCheckStatus.FINISHED
            assert nsset_mock.mock_calls == [call(self.nsset.nsset_id)]
            assert contact_mock.mock_calls == [
                call(self.contact1.contact_id),
                call(self.contact2.contact_id),
            ]
            assert email_mock.mock_calls == [
                call(
                    Email(
                        recipient="rimmer@example.com",
                        subject_template=MessageTemplate.DNSCHECK_SUBJECT,
                        body_template=MessageTemplate.DNSCHECK_BODY,
                        type=EmailType.DNSCHECK,
                        context={
                            "total": {TestResultLevel.CRITICAL: 1},
                            "nsset_handle": self.nsset.nsset_handle,
                            "nsset_registrar": self.nsset.sponsoring_registrar,
                            "started_at": self.nsset_check.started_at.isoformat(),
                            "finished_at": self.nsset_check.finished_at.isoformat(),
                            "bad_domains": [
                                {
                                    "fqdn": self.domain_check1.domain,
                                    "url": f"http://localhost/en/result/{self.domain_check1.test_id}",
                                }
                            ],
                        },
                    ),
                    references=[
                        Reference(type=ReferenceType.CONTACT, value=self.contact1.contact_id),
                        Reference(type=ReferenceType.NSSET, value=self.nsset.nsset_id),
                    ],
                ),
                call(
                    Email(
                        recipient="smeghead@example.com",
                        subject_template=MessageTemplate.DNSCHECK_SUBJECT,
                        body_template=MessageTemplate.DNSCHECK_BODY,
                        type=EmailType.DNSCHECK,
                        context={
                            "total": {TestResultLevel.CRITICAL.value: 1},
                            "nsset_handle": self.nsset.nsset_handle,
                            "nsset_registrar": self.nsset.sponsoring_registrar,
                            "started_at": self.nsset_check.started_at.isoformat(),
                            "finished_at": self.nsset_check.finished_at.isoformat(),
                            "bad_domains": [
                                {
                                    "fqdn": self.domain_check1.domain,
                                    "url": f"http://localhost/en/result/{self.domain_check1.test_id}",
                                }
                            ],
                        },
                    ),
                    references=[
                        Reference(type=ReferenceType.CONTACT, value=self.contact2.contact_id),
                        Reference(type=ReferenceType.NSSET, value=self.nsset.nsset_id),
                    ],
                ),
            ]
        logs.check(
            (
                "dnscheck.process.nsset_checks",
                "INFO",
                f"Sending email notifications about nsset {self.nsset.nsset_id}.",
            ),
            (
                "dnscheck.process.nsset_checks",
                "INFO",
                f"Message M1 to {self.contact1.emails[0]} (contact {self.contact1.contact_handle}) about nsset {self.nsset.nsset_id} was sent.",
            ),
            (
                "dnscheck.process.nsset_checks",
                "INFO",
                f"Message M2 to {self.contact2.emails[0]} (contact {self.contact2.contact_handle}) about nsset {self.nsset.nsset_id} was sent.",
            ),
        )

    @override_settings(messenger_netloc="localhost:9999")
    async def test_level_warning(self, db, mocker: MockFixture):
        self.nsset.technical_check_level = 3
        self.nsset.technical_contacts = [ContactRef(id=self.contact1.contact_id, handle=self.contact1.contact_handle)]
        mock_nsset = mocker.patch.object(APP.nsset_client, "get_nsset_info", return_value=self.nsset)
        mock_contact = mocker.patch.object(APP.contact_client, "get_contact_info", side_effect=[self.contact1])
        mock_email = mocker.patch.object(APP.email_client, "send", return_value="M1")

        with atomic() as sess:
            sess.delete(self.domain_check1)

        await nsset_checks.finish_one(1)

        with atomic() as sess:
            sess.add_all([self.nsset_check, self.domain_check2, self.domain_check3])

            assert self.nsset_check.status == NssetCheckStatus.FINISHED
            assert mock_nsset.mock_calls == [call(self.nsset.nsset_id)]
            assert mock_contact.mock_calls == [call(self.contact1.contact_id)]
            assert mock_email.mock_calls == [
                call(
                    Email(
                        recipient="rimmer@example.com",
                        subject_template=MessageTemplate.DNSCHECK_SUBJECT,
                        body_template=MessageTemplate.DNSCHECK_BODY,
                        type=EmailType.DNSCHECK,
                        context={
                            "total": {
                                TestResultLevel.CRITICAL.value: 0,
                                TestResultLevel.ERROR.value: 2,
                                TestResultLevel.WARNING.value: 3,
                            },
                            "nsset_handle": self.nsset.nsset_handle,
                            "nsset_registrar": self.nsset.sponsoring_registrar,
                            "started_at": self.nsset_check.started_at.isoformat(),
                            "finished_at": self.nsset_check.finished_at.isoformat(),
                            "bad_domains": [
                                {"fqdn": dc.domain, "url": f"http://localhost/en/result/{dc.test_id}"}
                                for dc in [self.domain_check2, self.domain_check3]
                            ],
                        },
                    ),
                    references=[
                        Reference(type=ReferenceType.CONTACT, value=self.contact1.contact_id),
                        Reference(type=ReferenceType.NSSET, value=self.nsset.nsset_id),
                    ],
                )
            ]

    @pytest.mark.parametrize("max_notified_domains,expected_domains", ((1, 1), (2, 2), (3, 3), (4, 3)))
    async def test_max_notified_domains(
        self, db, mocker: MockFixture, max_notified_domains: int, expected_domains: int
    ):
        with override_settings(messenger_netloc="localhost:9999", max_notified_domains=max_notified_domains):
            self.nsset.technical_check_level = 3
            self.nsset.technical_contacts = [
                ContactRef(id=self.contact1.contact_id, handle=self.contact1.contact_handle)
            ]
            mock_nsset = mocker.patch.object(APP.nsset_client, "get_nsset_info", return_value=self.nsset)
            mock_contact = mocker.patch.object(APP.contact_client, "get_contact_info", side_effect=[self.contact1])
            mock_email = mocker.patch.object(APP.email_client, "send", side_effect=["M1"])

            await nsset_checks.finish_one(1)

            with atomic() as sess:
                sess.add_all([self.nsset_check, self.domain_check1])

                assert self.nsset_check.status == NssetCheckStatus.FINISHED
                assert mock_nsset.mock_calls == [call(self.nsset.nsset_id)]
                assert mock_contact.mock_calls == [call(self.contact1.contact_id)]
                assert len(mock_email.mock_calls[0].args[0].context["bad_domains"]) == expected_domains

    @override_settings(messenger_netloc="localhost:9999", zonemaster_language="cs")
    async def test_zonemaster_language(self, db, mocker: MockFixture):
        self.nsset.technical_check_level = 1
        nsset_mock = mocker.patch.object(APP.nsset_client, "get_nsset_info", return_value=self.nsset)
        contact_mock = mocker.patch.object(
            APP.contact_client, "get_contact_info", side_effect=[self.contact1, self.contact2]
        )
        email_mock = mocker.patch.object(APP.email_client, "send", side_effect=["M1", "M2"])

        await nsset_checks.finish_one(1)

        with atomic() as sess:
            sess.add_all([self.nsset_check, self.domain_check1])

            assert self.nsset_check.status == NssetCheckStatus.FINISHED
            assert nsset_mock.mock_calls == [call(self.nsset.nsset_id)]
            assert contact_mock.mock_calls == [
                call(self.contact1.contact_id),
                call(self.contact2.contact_id),
            ]
            bad_domains = email_mock.mock_calls[0].args[0].context["bad_domains"]
            assert bad_domains[0]["url"] == f"http://localhost/cs/result/{self.domain_check1.test_id}"

    @override_settings(messenger_netloc="localhost:9999")
    async def test_send_failed(self, db, mocker: MockFixture):
        self.nsset.technical_check_level = 1
        nsset_mock = mocker.patch.object(APP.nsset_client, "get_nsset_info", return_value=self.nsset)
        self.contact1.emails = ["invalid@example.com", "valid@example.com"]
        contact_mock = mocker.patch.object(
            APP.contact_client, "get_contact_info", side_effect=[self.contact1, self.contact2]
        )
        email_mock = mocker.patch.object(APP.email_client, "send", side_effect=[SendFailed, "M1", "M2"])

        with LogCapture() as logs:
            await nsset_checks.finish_one(1)

        with atomic() as sess:
            sess.add_all([self.nsset_check, self.domain_check1])

            assert self.nsset_check.status == NssetCheckStatus.FINISHED
            assert nsset_mock.mock_calls == [call(self.nsset.nsset_id)]
            assert contact_mock.mock_calls == [
                call(self.contact1.contact_id),
                call(self.contact2.contact_id),
            ]
            assert email_mock.call_count == 3

        logs.check(
            (
                "dnscheck.process.nsset_checks",
                "INFO",
                f"Sending email notifications about nsset {self.nsset.nsset_id}.",
            ),
            (
                "dnscheck.process.nsset_checks",
                "WARNING",
                f"Message to {self.contact1.emails[0]} (contact {self.contact1.contact_handle}) "
                f"about nsset {self.nsset.nsset_id} failed. Skipping.",
            ),
            (
                "dnscheck.process.nsset_checks",
                "INFO",
                f"Message M1 to {self.contact1.emails[1]} (contact {self.contact1.contact_handle}) about nsset {self.nsset.nsset_id} was sent.",
            ),
            (
                "dnscheck.process.nsset_checks",
                "INFO",
                f"Message M2 to {self.contact2.emails[0]} (contact {self.contact2.contact_handle}) about nsset {self.nsset.nsset_id} was sent.",
            ),
        )

    async def test_nsset_does_not_exist(self, db, mocker: MockFixture):
        self.nsset.technical_check_level = 1
        nsset_mock = mocker.patch.object(APP.nsset_client, "get_nsset_info", side_effect=NssetDoesNotExist)

        with atomic() as sess:
            sess.add(self.nsset_check)
            old_nsset_id = self.nsset_check.nsset_id

        with LogCapture() as logs:
            await nsset_checks.finish_one(1)

        with atomic() as sess:
            assert [] == sess.execute(select(NssetCheck).where(NssetCheck.nsset_id == old_nsset_id)).all()
            assert nsset_mock.mock_calls == [call(old_nsset_id)]

        logs.check(
            (
                "dnscheck.process.nsset_checks",
                "WARNING",
                f"Nsset {old_nsset_id} does not exist anymore. Aborting nsset check.",
            )
        )

    @override_settings(messenger_netloc="localhost:9999")
    async def test_contact_does_not_exist(self, db, mocker: MockFixture):
        self.nsset.technical_check_level = 1
        nsset_mock = mocker.patch.object(APP.nsset_client, "get_nsset_info", return_value=self.nsset)
        contact_mock = mocker.patch.object(
            APP.contact_client, "get_contact_info", side_effect=[ContactDoesNotExist, self.contact2]
        )
        email_mock = mocker.patch.object(APP.email_client, "send", side_effect=["M2"])

        with LogCapture() as logs:
            await nsset_checks.finish_one(1)

        with atomic() as sess:
            sess.add_all([self.nsset_check, self.domain_check1])

            assert self.nsset_check.status == NssetCheckStatus.FINISHED
            assert nsset_mock.mock_calls == [call(self.nsset.nsset_id)]
            assert contact_mock.mock_calls == [
                call(self.contact1.contact_id),
                call(self.contact2.contact_id),
            ]
            assert email_mock.call_count == 1

        logs.check(
            (
                "dnscheck.process.nsset_checks",
                "INFO",
                f"Sending email notifications about nsset {self.nsset.nsset_id}.",
            ),
            (
                "dnscheck.process.nsset_checks",
                "WARNING",
                f"Contact {self.contact1.contact_id} does not exist. Skipping.",
            ),
            (
                "dnscheck.process.nsset_checks",
                "INFO",
                f"Message M2 to {self.contact2.emails[0]} (contact {self.contact2.contact_handle}) about nsset {self.nsset.nsset_id} was sent.",
            ),
        )
