from threading import Event

import click
import pytest
from click.testing import CliRunner
from testfixtures import LogCapture

from dnscheck.settings import APP
from dnscheck.utils import StopProcess, async_to_sync, defer_errors, run_unless_stopped, verbosity_options


def test_async_to_sync_synchronous_context():
    async def func(x: int) -> int:
        return x

    assert async_to_sync(func)(42) == 42


async def _ident(v):
    return v


async def _raise():
    0 / 0


async def test_run_unless_stopped_not_set():
    stop_event = Event()

    assert await run_unless_stopped(_ident, stop_event=stop_event)(42) == 42


async def test_run_unless_stopped_set():
    stop_event = Event()
    stop_event.set()

    with pytest.raises(StopProcess):
        await run_unless_stopped(_ident, stop_event=stop_event)(42)


async def test_defer_errors_ok():
    with LogCapture() as logs:
        assert await defer_errors(_ident)(42) == 42

    logs.check()


async def test_defer_errors_error():
    with LogCapture() as logs:
        assert await defer_errors(_raise)() is None

    logs.check(
        ("dnscheck.utils", "ERROR", "Unexpected error: ZeroDivisionError('division by zero')"),
    )


def test_verbosity_options():
    @click.command()
    @verbosity_options
    def cli():
        print(APP.verbosity)

    runner = CliRunner()

    assert runner.invoke(cli, ["-q"]).output == "0\n"
    assert runner.invoke(cli, ["--quiet"]).output == "0\n"
    assert runner.invoke(cli, ["--silent"]).output == "0\n"
    assert runner.invoke(cli, []).output == "1\n"
    assert runner.invoke(cli, ["-v"]).output == "2\n"
    assert runner.invoke(cli, ["--verbose"]).output == "2\n"
    assert runner.invoke(cli, ["-vv"]).output == "3\n"
    assert runner.invoke(cli, ["--verbose", "--verbose"]).output == "3\n"
    assert runner.invoke(cli, ["-qvv"]).output == "0\n"
    assert runner.invoke(cli, ["--quiet", "--verbose", "--verbose"]).output == "0\n"
    assert runner.invoke(cli, ["--silent", "--verbose", "--verbose"]).output == "0\n"
