"""DNScheck utils."""

import asyncio
import logging
from collections.abc import Callable, Coroutine
from functools import wraps
from threading import Event
from typing import Any, Concatenate, ParamSpec, TypeVar

import click
from setapp import Verbosity

from dnscheck.settings import APP

P = ParamSpec("P")
T = TypeVar("T")

_LOGGER = logging.getLogger(__name__)


def async_to_sync(func: Callable[P, Coroutine[None, None, T]]) -> Callable[P, T]:
    """Run async function.

    This function can only be called from synchronous context.
    If you're in asynchronous context, use `await` directly.
    """

    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        return asyncio.run(func(*args, **kwargs))

    return wrapper


class StopProcess(Exception):
    """Stop processing."""


def run_unless_stopped(func: Callable[P, T], *, stop_event: Event) -> Callable[P, T]:
    """Run function unless stop event has been triggered."""

    @wraps(func)
    def _wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        if stop_event.is_set():
            raise StopProcess
        else:
            return func(*args, **kwargs)

    return _wrapper


def defer_errors(func: Callable[P, Coroutine[None, None, T]]) -> Callable[P, Coroutine[None, None, T | None]]:
    """Log errors and continue running as if nothing happened."""

    @wraps(func)
    async def _wrapper(*args: P.args, **kwargs: P.kwargs) -> T | None:
        try:
            return await func(*args, **kwargs)
        except Exception as err:
            _LOGGER.exception("Unexpected error: %r", err)
            return None

    return _wrapper


def verbosity_options(func: Callable[P, T]) -> Callable[Concatenate[int, bool, P], T]:
    """Add --quiet and --verbose options to CLI command."""

    @click.option("-v", "--verbose", count=True, help="Increase verbosity by 1. Can be added multiple times.")
    @click.option("-q", "--quiet", "--silent", is_flag=True, help="Silence all output. Overrides --verbose.")
    @wraps(func)
    def _wrapper(verbose: int, quiet: bool, *args: P.args, **kwargs: P.kwargs) -> T:
        if quiet:
            APP.verbosity = Verbosity.SILENT
        elif verbose:
            APP.verbosity = Verbosity(min(Verbosity.DEBUG, verbose + 1))
        else:
            APP.verbosity = Verbosity.INFO
        return func(*args, **kwargs)

    return _wrapper


def print(*objects: Any, level: Verbosity = Verbosity.INFO, **kwargs: Any) -> None:
    """Print message if vebosity level is high enough."""
    if level <= APP.verbosity:
        APP.console.print(*objects, **kwargs)
